
# Copyright (C) 2021 Frank Sauerburger

"""
Toolkit to inspect the header and table directory.
"""

import sys
import struct
from colorama import Fore, Style

__version__ = "0.0.0"  # Also change in setup.py

class TrueType:
    """Representation of a TrueType font"""

    def __init__(self, file_name):
        """Parse TTF file"""
        self.directory = {}
        self.tables = {}

        with open(file_name, "rb", buffering=4096) as font_file:
            def get(fmt):
                size = struct.calcsize(fmt)
                data = font_file.read(size)
                return struct.unpack(fmt, data)

            self.v_major, self.v_minor = get(">HH")
            self.num_entries, self.search_range = get(">HH")
            self.entry_sel, self.range_shift = get(">HH")

            for _ in range(self.num_entries):
                tag, *pos = get(">IIII")
                tag = tag.to_bytes(4, "big")
                self.directory[tag] = tuple(pos)

            for tag, pos in self.directory.items():
                _, offset, length = pos
                font_file.seek(offset)
                self.tables[tag] = font_file.read(length)

    def checksum(self, tag):
        """Compute checksum of binary table content"""
        table = self.tables[tag]
        sum_ = 0
        pad = len(table) % 4
        if pad != 0:
            table += b"\x00" * (4 - pad)
        for i, (word, ) in enumerate(struct.iter_unpack(">I", table)):
            if i == 2 and tag == b"head":
                # Skip checksum adjust
                continue
            sum_ += word
        return sum_ % (1 << 32)

    def get_triple(self):
        """Return required value for search_range, entry_sel and range_shift"""
        search_range = 1
        entry_sel = 0
        while search_range <= self.num_entries:
            search_range <<= 1
            entry_sel += 1
        entry_sel = max(entry_sel - 1, 0)
        search_range = (search_range >> 1) * 16
        range_shift = self.num_entries * 16 -  search_range

        return search_range, entry_sel, range_shift

def fassert(is_value, should, message):
    """User interface assert"""
    fmt = "{message:30s} {is_value:20}  "
    if is_value == should:
        resolution = Fore.GREEN + "ok" + Fore.RESET
    else:
        resolution = Fore.RED + Style.BRIGHT + str(should) + Style.RESET_ALL

    print(fmt.format(is_value=is_value, message=message) + resolution,
        file=sys.stderr
    )


def check(file_name):
    """Run a check for a font"""
    font = TrueType(file_name)

    fassert(font.v_major, 1, "Major TTF version")
    fassert(font.v_minor, 0, "Minor TTF version")

    fassert(font.num_entries, font.num_entries, "Number of tables")

    search_range, entry_sel, range_shift = font.get_triple()
    fassert(font.search_range, search_range, "Search range")
    fassert(font.entry_sel, entry_sel, "Entry selection")
    fassert(font.range_shift, range_shift, "Range shift")

    for tag, table in font.tables.items():
        fassert(font.directory[tag][0], font.checksum(tag),
                f"Checksum {tag!r} table")
